// cbi and sbi are standard (AVR) methods for setting, 
// or clearing, bits in PORT (and other) variables.
#ifndef cbi
#define cbi(sfr, bit) (_SFR_BYTE(sfr) &= ~_BV(bit))
#endif
#ifndef sbi
#define sbi(sfr, bit) (_SFR_BYTE(sfr) |= _BV(bit))
#endif 

//---------------------------------------------------------------------- 
void EnableADC()
  { // This is probably not needed but ....
    // set system clock devisor to 128
    // 16 MHz / 128 = 125 KHz, inside the desired 50-200 KHz range.
    sbi(ADCSRA, ADPS2);                            // bit 2 of ADCSRA, system clock devisor
    sbi(ADCSRA, ADPS1);                            // bit 1 of ADCSRA, system clock devisor
    sbi(ADCSRA, ADPS0);                            // bit 0 of ADCSRA, system clock devisor
    cbi(ADCSRA, ADATE);                            // bit 5 of ADCSRA, disable auto trigger mode
    sbi(ADCSRA, ADEN);                             // bit 7 of ADCSRA, enable ADC 
  } 

//---------------------------------------------------------------------- 
void Read_Calibration_Data()
 { // This is more or less taken straight from the 
   // EEPROM_TempSensor_Calibration_Constants program
   byte i=0, j=0;
   word  TempWord;
   char c;
   char sign;
   
   // detect a virgin device --- well at least try
   if ((EEPROM.read(EEminutes)==0xFF) &&
       (EEPROM.read(EEminutes+1)==0xFF))
	   TestData1();
 
   // Get Degree Offset
   TempWord= EEPROM.read(EEoffset)<<8;
   TempWord= (TempWord + EEPROM.read(EEoffset +1));
   // Serial.print (F("Got Here: Degree Offset EEPROM word: "));
   // Serial.println     (TempWord, HEX);
   // get the sign bit
   sign= 1;
   // Serial.print    (F("High Nibble: "));
   // Serial.println    (TempWord>>12,BIN);
   if ((TempWord>>12) == B1000) sign= -1;
   // Serial.print (F("Sign: "));
   // Serial.println     (sign,DEC);
   // strip the sign bit
   TempWord=(TempWord<<1)>>1;
   // Serial.print (F("Degree Offset EEPROM word stripped: "));
   // Serial.println     (TempWord, HEX);
   // now we need to convert it to a fraction
   DegreeOffset=(float(TempWord)/float(CovrtFactorO)) * sign;

   // Get the RefVoltage factor;
   TempWord= EEPROM.read(EErefvolt)<<8;
   TempWord= TempWord + EEPROM.read(EErefvolt +1);
   // now we need to convert it to a fraction
   RefVoltage = float(TempWord)/CovrtFactorV ;

   // ID String -------------------------------------
   i=0; c=1;
   while (c!=0,i< EEidsize)
    {c=EEPROM.read(EEidtring + i);         // read the ID string
     IdString [i++]=c; 
    } 
   IdString [EEidsize]=0;                  // just in case

   // Set MinuteTarget from default minutes
   MinuteTarget=(EEPROM.read(EEminutes)<<8) + EEPROM.read(EEminutes +1);
   if (MinuteTarget<1)MinuteTarget=1;
   
   // this is a bit flag to indicate when the current constants
   // in memory are different from those stored in working storage 
   newflg=0; 
   
 }
 
//----------------------------------------------------------------------  
void Write_Calibration_Data()
  { word i=0;
    // float saveflt;
    word  TempWord;
    char c=-1;
    byte sign;

    // New location
    if ((newflg & B00000001) == B00000001)
      {// Serial.println ("Got Here: Write_Calibration_Data, Location");  
        for (i=0;i<EEidsize; i++) {EEPROM.write((EEidtring +i), IdString[i]);}
      }

    // New Degree Offset
    if ((newflg & B00000010) == B00000010)
      { // Serial.println (F("Got Here: Write_Calibration_Data, Offset"));
        // Serial.print   (F("Degree Offset: "));
        // Serial.println   (DegreeOffset);
        sign = 0;
        if (DegreeOffset <0) sign = B10000000;
        TempWord=word(abs(DegreeOffset) * CovrtFactorO);   
        // Serial.print   (F("Degree Offset as word: "));
        // Serial.println   (TempWord);
        // Serial.print   (F("Sign: "));
        // Serial.println   (sign);
        EEPROM.write(EEoffset   , highByte(TempWord)|sign);
        EEPROM.write(EEoffset +1, lowByte (TempWord));
      }
    // New Reference Voltage
    if ((newflg & B00000100) == B00000100)
      { //Serial.println ("Got Here: Write_Calibration_Data, Reference Voltage");
        TempWord = word(RefVoltage * CovrtFactorV);
        EEPROM.write(EErefvolt   , highByte(TempWord));
        EEPROM.write(EErefvolt +1, lowByte (TempWord));
        // Note: 
        // When the EEPROM data is read
        // Covrt2Fahrenheit is calculated from RefVoltage
      }
      
    // New MinuteTarget
    if ((newflg & B00010000) == B00010000)
      {// Serial.println ("Got Here: Write_Calibration_Data, MinuteTarget");
        EEPROM.write(EEminutes   , highByte(MinuteTarget));
        EEPROM.write(EEminutes +1, lowByte (MinuteTarget));
      }

     // write the unused word(s)
     i=EEunused0;
      while (i<EEidtring)
       {if (EEPROM.read(i)!= 0xFF) EEPROM.write(i,  0xFF);
        i++;
       } 
      
    // clear the EEMODE flag
    EEmodeFlagClear();

   // Serial.println ("Got Here: Write_Calibration_Data, Read");
   // note newflg is reset by Read_Calibration_Data
   Read_Calibration_Data();
   // PrintOKStr(); is sent by Report_Reset
   Report_Reset();
  } 
 
//-----------------------------------------------------
void ClearStorage()
 { // this is used to clear/erase the EEPROM data storage (except for constant areas)  
   word addr;
   byte b;
   for (addr=(StorageBegin); addr<StorageBackup; addr++)
     { if (EEPROM.read(addr) != 0xFF) EEPROM.write(addr,0xFF);
     } // note: each byte requires 6-8 machine cycles
   PrintOKStr ();
 }     
 
//-----------------------------------------------------
 void EEmodeFlagSet()
  { // toggle the flag forthe next run to write to EEPROM
    // does not affect current run
    // Here is the thing. We have setup wearleveling for our EEPROM data storage
    // but modeflag gets hit twice for every EEPROM run. So it will wearout
    // long before the bulk of the storage. We could just increment the byte and
    // look for odd or even values but that would continuelly toggle the low bit
    // wearing it out before there rest. As it turns out it is writing a zero to 
    // a bit that wears then out. So we want to minimize the zero bit writes.
    // We are going to move a zero bit right to left. This extends our life by a
    // factor of eight. At that point you need to swap the backup and working
    // data locations by changing the EEPROM address locations and reprograming 
    // the Arduino. That would double the life (2 * 8 = 16).
    //
    byte flag;
    if(EEmodeFlagTF()==false)
      { // Serial.println (F("Got here: EEmodeFlagToggle, make not equal"));
        // make them not equal
        // shift left and add a one to the right
        flag=EEPROM.read(EEflag);
        // Serial.println(flag);    
        flag=(flag<<1)+1;
        // Serial.println(flag);    
        // if we have all ones start over again at the right  
        if (flag==B11111111) flag=B11111110;
        // Serial.println(flag);    
        // now save it
        EEPROM.write(EEflag,flag);
      }
    PrintOKStr();
  }

//-----------------------------------------------------
 void EEmodeFlagClear()
  { byte flag;
    if(EEmodeFlagTF())
      { // Serial.println (F("Got here: EEmodeFlagClear"));
        // make them equal
        flag=EEPROM.read(EEflag);
        EEPROM.write(EEmask,flag);
      }
    if (EepromMode==false) PrintOKStr();
  }

//----------------------------------------------------
boolean EEmodeFlagTF()
  { // returns true if EEmodeFlag is set
    byte flag,mask;
    flag=EEPROM.read(EEflag);
    mask=EEPROM.read(EEmask);
    if(flag==mask) return false;
    else return true;
  } 

//-----------------------------------------------------
void Check_EEPROM()
 { byte b[4],i;
   word w;
   word addr;
   
   EepromMode=false;
   if (EEmodeFlagTF())
     { // Serial.println ("got here: Check_EEPROM");
       EepromMode=true;
       // clear the flag
       EEmodeFlagClear();
       // disable serial reporting and debug mode
       ReportMode=false;
       DeBug=false;
       // we need to find the beginning of EEPROM that has not been used
       // we need at least four bytes to begin a new section.
       // so we need to find the first place where there are four bytes with FFh
       // zero our test pattern
       for (i=0; i<4; i++) b[i]=0;
       addr=StorageBegin;
       while ((addr<StorageBackup) && ((b[0]!=0xFF) || (b[1]!=0xFF) 
                                   || (b[2]!=0xFF) || (b[3]!=0xFF)))
         { b[0]=b[1];
           b[1]=b[2];
           b[2]=b[3];
           b[3]=EEPROM.read(addr++);
         }
       // did we read until the end ??
       if (addr >= StorageEnd) StorageMark=StorageBegin;
       // we found 4 bytes that have not been written to
       else StorageMark=addr-4;
       // in either case we clear the storage  
       ClearStorage();
       // mark the beginning
       EEPROM.write(StorageMark,0);
       EEPROM.write(StorageMark+1,0);
       // set then beginning and end of the current segment
       StorageIndex=StorageMark+2;
       StorageEnd=StorageBackup;
       // definitive notice of mode
       for (i=0; i < 30; i++) {QuickBlink(); delay (250);}
       // -------
       // while(true);  // stop here so we can check a memory dump
     }
 }

//-----------------------------------------------------
void Print_IdString()
 { // Serial.print(F("; "));                    // pefix 
   Serial.println(IdString);                 // print it
 }
 
//-----------------------------------------------------
void PrintTrueFalse(byte T) 
 { // used to report True or False for boolean Globals
   if(T == 0) Serial.println(F("False"));
   else  Serial.println(F("True"));
 }
     
//-----------------------------------------------------
void ReportStatus()
 { // report settings
   PrintSeperatorLine();
   Serial.print   (F(";    Report:\t"));
   PrintTrueFalse (ReportMode);
   Serial.print   (F(";     Debug:\t"));
   PrintTrueFalse (DeBug);
   Serial.print   (F(";       Raw:\t"));
   PrintTrueFalse (RtnRawRead);
   Serial.print   (F(";Fahrenheit:\t"));
   PrintTrueFalse (RtnFahrenh);
   Serial.print   (F(";   Celsius:\t"));
   PrintTrueFalse (RtnCelsius);
   Serial.print   (F(";       Avr:\t"));
   PrintTrueFalse (RtnAvrRead);
   Serial.print   (F(";     Round:\t"));
   PrintTrueFalse (RoundMode);
   
   Serial.print   (F(";   Minutes:\t"));
   Serial.println (MinuteTarget, DEC);
   Serial.print   (F(";   Voltage:\t"));
   Serial.println (RefVoltage, 4);

   // single sensor, no prefix needed
   Serial.print   (F("; Sensor ID:\t"));
   Print_IdString ();
   Serial.print   (F(";    Offset:\t"));
   Serial.println (DegreeOffset, 4);

   if (newflg != 0) 
     Serial.println (F(";Parameters not saved"));
   if(EEmodeFlagTF()) 
     Serial.println (F(";EEPROM Mode Flag Set"));
      
   PrintSeperatorLine();
   
 }
//-----------------------------------------------------------------------------------------
//    ---- See Below for the function the read the External LM34DZ Temperature Sensor ----
//-----------------------------------------------------------------------------------------
void AvrTemperature()
  { // Read AVR internal Temperature Sensor and Convert to Celsius and Degrees
    //
    unsigned long RawSum=0;                         // used to sum samples for averaging 
    word  RawTemp=0;                                // used to accumalate 10 bit ADC readings 
    word  test=0;                                   // used to count samples
    word  RawReading;                               // averaged reading
    float RawVoltage;
    float DegreesKelvin;
    
    // turn on internal reference, right-shift ADC buffer,ADC channel = Avr temperature
    // B11000000: AMUX Voltage reference  = Internal 1.1 volt        (bits 7,6)
    // B00000000: AMUX Right Shift ADC Buffer                        (bit 5)
    // B00001000: AMUX Input Source= internal temperature sensor     (bits 3,2,1,0)
    ADMUX = B11001000;
    delay(10);    			            // wait for the things to stabilize
    
    while (test++ < 1024 )                          // oversampling loop (for averaging) 
      { ADCSRA |= _BV(ADSC); 	                    // start the conversion                                
        while (bit_is_set(ADCSRA, ADSC));           // ADSC cleared when complete
        RawTemp = (ADCL | (ADCH << 8));             // collect the reading
        RawSum += RawTemp;                          // add it to out total
      }
    RawReading = (RawSum)>>10;                      // averag the 1024 readings
    //---- Convert  It------
    // Convert to Raw Voltage
    RawVoltage = (float(RawReading)/1024)* RefVoltage;
    // Convert to Temperature in degrees Kelvin
    DegreesKelvin= RawVoltage * 1000;
    // Convert to Celsius
    AvrCelsius= DegreesKelvin - 273;
    // Convert to Fahrenheit
    AVRFahrenheit=(AvrCelsius*1.8)+32;
    //---- Print It------
    Serial.print     (";");
    Serial.print     (RawReading);
    Serial.print     (char(9));
    // Serial.print     (RawVoltage);
    // Serial.print     (char(9));
    // Serial.print     (DegreesKelvin);
    // Serial.print     (char(9));
    Serial.print     (AvrCelsius);
    Serial.print     (char(9));
    Serial.print     (AVRFahrenheit);
    Serial.print     (char(9));
    Serial.println (F("ARV"));
    
  }

//-----------------------------------------------------------------------------------------
//    ---- Below is the function the read the External LM34DZ Temperature Sensor ----
//-----------------------------------------------------------------------------------------
void ReadRawTempA1()
  { // Read ADC for Pin A1 (connected to LM34DZ Temperature Sensor)
    // Cycle time is aproximately 124 miliseconds
    unsigned long RawSum=0;                         // used to sum samples for averaging 
    word RawTemp=0;                                 // used to accumalate 10 bit ADC readings 
    word test=0;                                    // used to count samples
    
    // Just in case ...
    // The INPUT mode explicitly disables the internal pullup resistors.
    pinMode(A1,INPUT);
    
    // turn on internal reference, right-shift ADC buffer,ADC channel = ADC1 (pin A1)
    // B11000000: AMUX Voltage reference  = Internal 1.1 volt        (bits 7,6)
    // B00000000: AMUX Right Shift ADC Buffer                        (bit 5)
    // B00000001: AMUX Input Source= pin A1                          (bits 3,2,1,0)
    ADMUX = B11000001;
    delay(10);    			            // wait for the things to stabilize
    
    while (test++ < 1024 )                          // oversampling loop (for averaging) 
      { ADCSRA |= _BV(ADSC); 	                    // start the conversion                                
        while (bit_is_set(ADCSRA, ADSC));           // ADSC cleared when complete
        RawTemp = (ADCL | (ADCH << 8));             // collect the reading
        RawSum += RawTemp;                          // add it to out total
      }
    Accumalator += ((RawSum)>>10);                  // averag and add to Accumalator 
                                                    // LSBs are lost 
    CycleCount++;                                   // used by functions to average readings 
  }

//-----------------------------------------------------------------------------------------
void Convert(word RawReading)
  { // converts Raw Reading to Celsius and Fahrenheit
    // New plan: offset will only be used for minor correction
    // SCALE is actual voltage that is supposed to be 1.1 but reads 1.067
    // This temperature sensor reports in Fahrenheit 1 milivolt per degree
    // We need the correct voltage !!!
    // Covrt2Fahrenheit=1.067;
    Fahrenheit = (float(RawReading)/1024)* RefVoltage *100;
    Fahrenheit = Fahrenheit + DegreeOffset;
    if (RoundMode) Fahrenheit= nearesthalf(Fahrenheit);
    // Consistancy is next to godliness.
    // We are working in Fahrenheit. 
    // In our case Celsius is a function of Fahrenheit.
    // Thus we always complete our Fahrenheit cals first.
    // That inludes rounding.
    Celsius   = (Fahrenheit-32)/1.8;
    if (RoundMode) Celsius=nearestquater(Celsius);
  }

//-----------------------------------------------------
float nearestquater (float ValueIn)
  { // Return value rounded to nearest quater (0.25)
    byte sign=1;
    float fraction;
    if (ValueIn , 0) 
      {
      sign = -1; 
      ValueIn=abs(ValueIn);
      }
    fraction =ValueIn-long(ValueIn);
    if      (fraction >= 0.875) fraction=1.00; 
    else if (fraction >= 0.625) fraction=0.75;
    else if (fraction >= 0.375) fraction=0.50;
    else if (fraction >= 0.125) fraction=0.25;
    else                        fraction=0;
    // Serial.println (ValueIn);
    // Serial.println (ValueIn-long(ValueIn));
    // Serial.println (fraction);
    // Serial.println (sign);
    return (long(ValueIn)+fraction) * sign;
  }
//-----------------------------------------------------
float nearesthalf (float ValueIn)
  { // Return value rounded to nearest half (0.50)
    byte sign=1;
    float fraction;
    if (ValueIn , 0) 
      {
      sign = -1; 
      ValueIn=abs(ValueIn);
      }
    fraction =ValueIn-long(ValueIn);
    if      (fraction >= 0.750) fraction=1.00; 
    else if (fraction >= 0.250) fraction=0.50;
    else                        fraction=0;
    return (long(ValueIn)+fraction) * sign;
  }
//-----------------------------------------------------
void Report()
 { word AvgSumRead;
   AvgSumRead = Accumalator/CycleCount;
   if (EepromMode) Report2EEPROM(AvgSumRead); 
   else if (ReportMode)
     {  Convert (AvgSumRead);
          Serial.print    ('0'); 
        if (RtnRawRead)
          { Serial.print  (AvgSumRead);
            Serial.print  (char(9));
          }
        if (RtnCelsius)
          { Serial.print  (Celsius,2);
            Serial.print  (char(9));
          }
        if (RtnFahrenh)
          { Serial.print  (Fahrenheit,2);
            Serial.print  (char(9));
          } 
        if (DeBug)
          { Serial.print  (CycleTime/CycleCount);
            Serial.print  (char(9));
            Serial.print  (CycleCount);
            Serial.print  (char(9));
            Serial.print  (millis()-RptStartTime);
            RptStartTime=millis();
          }
       Serial.println();
       // addin for AVR internal temperature line
       if (RtnAvrRead) AvrTemperature();
     }
   Accumalator   = 0;
   CycleCount    = 0;
   CycleTime     = 0;
   LastRead=AvgSumRead;
 }

//------------------------------------------------------
void QuickBlink()
 { // on the UNO 1 mullisecond will surfice
   // adjusted up to 3 for Nano
   digitalWrite(13, HIGH);     // turn on LED
   delay(3);       
   digitalWrite(13, LOW);     // turn off LED
 }

//------------------------------------------------------
void Report2EEPROM(word AvgSumRead)
  { // We are implimenting both data compression and wearleveling.
    // Our data is only 12 bits. Becuase we should never get a reading 
    // over 2047 in our high bit will always be zero.
    // We are going to use the top four bits to count consecutive equal
    // readings. In that manner we may be able to store 16 readings in
    // a singal word value.
    word makeword;
    QuickBlink();
    
    // Serial.println(F("Get here: Report2EEPROM")); 
    // send this string for testing: EC EE ST !! !!
    // we need to skip the firs pass because we have nothing to work with
    if (LastRead !=0)
      { if (LastRead == AvgSumRead) Consecutive++;
        if ((Consecutive == 15) || (LastRead != AvgSumRead))
          { // Serial.println(F("Get here: Report2EEPROM, write record"));
            // we are going to try two blinks everytime that there is a write
            makeword = (Consecutive <<12)+LastRead;
            EEPROM.write (StorageIndex++, highByte(makeword));
            EEPROM.write (StorageIndex++,  lowByte(makeword));
            Consecutive=0;
            // now we need to check our storage space
            if ((StorageEnd-StorageIndex)<2)
              { // folks there is Trouble in river city !
                if (StorageMark==StorageBegin) 
                            {prevcmd[1]=prevcmd[0]=cmd[1]=cmd[0]='S'; ShutDown();}
                if (StorageEnd ==StorageMark ) 
                            {prevcmd[1]=prevcmd[0]=cmd[1]=cmd[0]='S'; ShutDown();}
                // OK, start at the beginning ....
                StorageIndex=StorageBegin;
                StorageEnd=StorageMark;
              }
            delay (50);     // force a bit of a delay so both blinks can be seen  
            QuickBlink();
          }
      }
      LastRead == AvgSumRead;
  }  

//------------------------------------------------------  
void DumpStorage()  
 { // print data stored in eeprom
   byte b1, b2, c;
   word reading;
   word countreading=0;
   word countwords=0;
   boolean savemode;
   // save the current reporting mode
   savemode=ReportMode;
   // find the beginning of the data defined to be two zero bytes
       b1=1;
       b2=1;
       while ((StorageIndex<StorageBackup) && ((b1!=0) || (b2!=0)))
         { b1=b2;
           b2=EEPROM.read(StorageIndex++);
         }
       StorageEnd=StorageBackup;
       StorageMark=StorageIndex-2;
       // 4 high bits are the count, low 12 bits are the reading  
       PrintSeperatorLine();
       Serial.println (F("; Begin EEPROM data dump"));
       Serial.println (F("; Raw Reading\tCelsius\tFahrenheit"));
       while (((StorageEnd-StorageIndex)>=2) && ((b1 != 0xFF)||(b2 != 0xFF)))
        { b1=EEPROM.read(StorageIndex++);
          b2=EEPROM.read(StorageIndex++); 
          //----------------- debuggin code
          // Serial.print   ("; Location: ");
          // Serial.print   (StorageIndex);
          // Serial.print   (", ");
          // Serial.print   (b1,HEX);
          // Serial.print   (", ");
          // Serial.print   (b2,HEX);
          countwords++;
          // two bytes of FFh will mark the end
          if ((b1 != 0xFF) || (b1 != 0xFF))
            { Consecutive=b1>>4;
              reading= ((b1 & B00001111)<<8)+b2;
              Convert(reading);
              //----------------- debuggin code
              // Serial.print   (", ");
              // Serial.print   (Consecutive);
              // Serial.print   (", ");
              // Serial.print   (reading);
              // Serial.println();
              // while (Serial.available() ==0);
              // c=Serial.read();
              // the logic here is we need to print every reading at least once ...
              // that is when it is zero. When we subtract one from zero we get 255
              while (Consecutive<255)
               { countreading++;
                 Serial.print   (reading);
                 Serial.print   (char(9));
                 Serial.print   (Celsius,2);
                 Serial.print   (char(9));
                 Serial.print   (Fahrenheit,2);
                 Serial.println ();
                 Consecutive--;
                 //----------------- debuggin code
                 // Serial.print   (Consecutive);
                 // Serial.print   (", ");
                 // Serial.println();
                 // while (Serial.available() ==0);
                 // c=Serial.read();
               }  
             // now check the addresses
             if ((StorageEnd-StorageIndex)<2)
               { if (StorageMark != StorageBegin)
                   { StorageEnd  = StorageMark;
                     StorageMark = StorageBegin;
                     StorageIndex = StorageMark +2;
                   }
               }
            }    
        }
       PrintSeperatorLine(); 
       Serial.println (F("; End EEPROM data dump"));
       Serial.print   (F(";     Readings:\t"));
       Serial.println   (countreading, DEC);
       Serial.print   (F(";Storage Words:\t"));
       Serial.println   (countwords, DEC);
       PrintSeperatorLine();
   // restore the current reporting mode
   ReportMode=savemode;
 }
//---------------------------------------------------------------------- 
void Responce (char str[])
  {if (EepromMode == false)
     { // we do not want to get hung up 
       // this just serves to reduce command responce memory usage a bit
       // trying to wrtie to seomthing that is not connected
       Serial.print (F("; "));
       Serial.print (cmd); 
       Serial.print (F(" "));
       Serial.println (str);
     }
  } 
   
//---------------------------------------------------------------------- 
void PrintOKStr ()
 { // command was accepted and processed
   // this just serves to reduce command responce memory usage a bit
   Responce ("OK");
 }  

//---------------------------------------------------------------------- 
void PrintNotRecognized()
 { // command was Not Recognized
   // this just serves to reduce command responce memory usage a bit
   Responce ("??");
 }  

//---------------------------------------------------------------------- 
void PrintNotImplemented()
 { // command was Not Recognized
   Responce ("XX");
 }  

//-----------------------------------------------------
void ShutDown()
  { // Note that no provision is made to wake up.
    // This is as close to shutdown as we can get.
    // Because of the inefficent voltage regulator this 
    // mode still draws a lot of power (about 10mA). 
    // A standard 9 volt battery may last about 16 hours.

    // Serial.println(prevcmd);
    if ((prevcmd[0]=='S') && (prevcmd[1]=='S'))
      { Serial.println (F("; SHUTDOWN"));
        // give device time to send string
        for (byte i=0; i< 25; i++)
          { QuickBlink();
            delay (100);
          }
        cbi(ADCSRA, ADEN);   // bit 7 of ADCSRA, disable ADC  
        noInterrupts();
        set_sleep_mode(SLEEP_MODE_PWR_DOWN);  
        sleep_enable();  
        sleep_mode();       // all execution should stop here
        while(0==0);        // endless loop (belts and suspenders)
      }
    else PrintOKStr();      // first time through only
  }

//---------------------------------------------------------------------- 
void software_Reset()
  { // Restarts program from beginning but 
    // does not reset the peripherals and registers
    // as we are not doing anything with the the
    // timers or peripherals or registers this
    // should be adequate (will not support updating)

    // Serial.println(prevcmd);
    if ((prevcmd[0]=='!') && (prevcmd[1]=='!'))
      { Serial.println (F("; RESETTING"));
        // give device time to send string
        delay (1000);
        asm volatile ("  jmp 0");
      }
    else PrintOKStr();      // first time through only
  }   

//---------------------------------------------------------------------- 
void SetRawReadMode()
  { // check for "T" or "F", true of false
    if      (cmd[1]=='T') {RtnRawRead = true;  PrintOKStr();}
    else if (cmd[1]=='F') {RtnRawRead = false; PrintOKStr();}
    else PrintNotRecognized();
  }

//---------------------------------------------------------------------- 
void SetCelsiusMode()
  { // check for "T" or "F", true of false
    if      (cmd[1]=='T') {RtnCelsius = true;  PrintOKStr();}
    else if (cmd[1]=='F') {RtnCelsius = false; PrintOKStr();}
    else PrintNotRecognized();
  }

//---------------------------------------------------------------------- 
void SetFahrenheitdMode()
  { // check for "T" or "F", true of false
    if      (cmd[1]=='T') {RtnFahrenh = true;  PrintOKStr();}
    else if (cmd[1]=='F') {RtnFahrenh = false; PrintOKStr();}
    else PrintNotRecognized();
  }

//---------------------------------------------------------------------- 
void SetReportMode()
  { // check for "T" or "F", true of false
    if      (cmd[1]=='T') {ReportMode = true;  PrintOKStr();}
    else if (cmd[1]=='F') {ReportMode = false; PrintOKStr();}
    else PrintNotRecognized();
  }

//---------------------------------------------------------------------- 
void ToggleDebugMode()
  { // toggle Debug mode
    if      (DeBug == true)  DeBug = false;
    else if (DeBug == false) DeBug = true;
    PrintOKStr();
  }

//---------------------------------------------------------------------- 
void SetAvrInternalMode()
  { // check for "T" or "F", true of false
    if      (cmd[1]=='T') {RtnAvrRead = true;  PrintOKStr();}
    else if (cmd[1]=='F') {RtnAvrRead = false; PrintOKStr();}
    else PrintNotRecognized();
  }

//---------------------------------------------------------------------- 
void ToggleRoundMode()
  { // check for "T" or "F", true of false
    if   (RoundMode) {RoundMode = false; PrintOKStr();}
    else             {RoundMode = true;  PrintOKStr();}
  }
  
//---------------------------------------------------------------------- 
void NewReportTime()
  { // set report Minutes
    if      (cmd[1]=='1') { MinuteTarget =    1; Report_Reset();} 
    else if (cmd[1]=='2') { MinuteTarget =    2; Report_Reset();} 
    else if (cmd[1]=='3') { MinuteTarget =    3; Report_Reset();}
    else if (cmd[1]=='4') { MinuteTarget =    4; Report_Reset();}
    else if (cmd[1]=='5') { MinuteTarget =    5; Report_Reset();}
    //---- the timings below have not been tested ---------------
    else if (cmd[1]=='6') { MinuteTarget =   10; Report_Reset();}
    else if (cmd[1]=='7') { MinuteTarget =   15; Report_Reset();}
    else if (cmd[1]=='8') { MinuteTarget =   20; Report_Reset();}
    else if (cmd[1]=='9') { MinuteTarget =   30; Report_Reset();}
    else if (cmd[1]=='0') { MinuteTarget =   60; Report_Reset();}
    else if (cmd[1]=='A') { MinuteTarget =  120; Report_Reset();}
    else if (cmd[1]=='B') { MinuteTarget =  240; Report_Reset();}
    else if (cmd[1]=='C') { MinuteTarget =  360; Report_Reset();}
    else if (cmd[1]=='D') { MinuteTarget =  480; Report_Reset();}
    else if (cmd[1]=='E') { MinuteTarget =  720; Report_Reset();}
    else if (cmd[1]=='F') { MinuteTarget = 1440; Report_Reset();}
    // max=86,400,000 milliseconds and that is why we use four byte variables
    
    else if (cmd[1]=='T') Serial.println(F("; TT XX")); // not implimented
    else PrintNotRecognized();                          // not recognized
  }

//----------------------------------------------------------------------  
void Report_Reset()
  { // this force the current data to be reported 
    // and reset our clock using the new time
    unsigned long SaveMe=SecondsTarget;
    PrintOKStr();
    Serial.println (F("; Report Timing reset"));
    // calculate seconds between report lines
    // SecondsTarget=MinuteTarget*SecondsMinute;
    // we have to "cast" the two word values or we will get a word value for the result
    SecondsTarget=(long(MinuteTarget)*long(SecondsMinute));
    
    if (SecondsTarget != SaveMe) newflg = newflg |  B00010000;
    // Serial.print (F("Got Here: report reset, milliseconds to wait= "));    
    // Serial.println (SecondsTarget);
    
    Accumalator = 0;                          // reset report parameters
    CycleCount  = 0;
    RptTrigger  = millis() + SecondsTarget;
    RptStartTime= millis();     
    // Serial.print (F("Got Here: report reset, trigger= "));
    // Serial.println (RptTrigger-millis());    
  }   

//---------------------------------------------------------------------- 
void NewIdString()
  { // New Location ID String
    // Serial.println("got here: NewIdString");
    // set time out to 5 seconds
    unsigned long timelimit = millis() + (5000);
    boolean timeout=false;
    char c= -1;
    byte n= 0;
    while ((c != 0) && (c != 10) && (c != 9) && (c != 13) && (n<EEidsize) && (timeout==false))
     { delay(10);
       c = Serial.read();
       if (c > 31) IdString[n++]=c;
       // check for timeout
       if (millis()>timelimit) timeout=true;
     }
   while (n<EEidsize) IdString[n++]=0;  
   if (timeout) Serial.println (F("; L: aborted due to timeout"));
   else
    { IdString[EEidsize]=0;  // make certain last charater is null
      newflg = newflg |  B00000001;
      // Serial.println (IdString);
      PrintOKStr();
    }
  DrainCmdTermiantors();
  }
  
//---------------------------------------------------------------------- 
void PrintDegreeOffsetEffect(float NewOffset)
  { // new offset must be in Degrees Fahrenheit 
    boolean SaveRoundMode;
    SaveRoundMode=RoundMode;
    
    while (CycleCount<50) ReadRawTempA1();
    DegreeOffset=0;
    RoundMode=false;
    Convert(Accumalator/CycleCount);
    DegreeOffset=NewOffset;
    Serial.print (F("; Offset(F):\t"));
    Serial.println (DegreeOffset);
    Serial.print (F(";Fahrenheit:\t"));
    Serial.println (Fahrenheit);
    Serial.print (F(";  Adjusted:\t"));
    Fahrenheit=Fahrenheit+NewOffset;
    if (SaveRoundMode) Fahrenheit=nearesthalf(Fahrenheit);
    Serial.println (Fahrenheit);
    //-----------------------------------------------
    Serial.print (F(";   Celsius:\t"));
    Serial.println (Celsius);
    Serial.print (F(";  Adjusted:\t"));
    Celsius=Celsius+(DegreeOffset/1.8000);
    if (SaveRoundMode) Celsius=nearestquater(Celsius);
    Serial.println (Celsius);
    newflg = newflg |  B00000010;
    RoundMode=SaveRoundMode;
    PrintOKStr();
  }

//---------------------------------------------------------------------- 
void ValueNotAccepted()
  { Serial.print("; ");
    Serial.print(cmd);
    Serial.print(" invalid/no input");
  }

//---------------------------------------------------------------------- 
void NewDegreeOffset()
  { // New Degree Offset
    float tempfloat=0;
    delay (2000);
    tempfloat=Serial.parseFloat();
    if (tempfloat!=0)
      { if (abs(tempfloat)<0.010) tempfloat = 0;
        PrintDegreeOffsetEffect(tempfloat);
        newflg = newflg |  B00000010;
      }
    else ValueNotAccepted();    
    DrainCmdTermiantors();
  }

//---------------------------------------------------------------------- 
void CalculateDegreeOffset(float tempfloat)
  { // calculate a new degree offset, TempF is Temperature in degrees Fahrenheit
    // get the current raw reading
    boolean SaveRoundMode;
    
    Serial.println (F("; Calculating new Degree offset ... ..."));
    while (CycleCount<50) ReadRawTempA1();
    // set the current offset to zero so that it
    // does not affect the Conversion
    DegreeOffset=0;
    SaveRoundMode=RoundMode;
    RoundMode=false;
    Convert(Accumalator/CycleCount);
    RoundMode=SaveRoundMode;
    PrintDegreeOffsetEffect(tempfloat-Fahrenheit);
  }  
  
//---------------------------------------------------------------------- 
void FahrenheitEquals()
 { // sets offset according to current reading and input Fahrenheit
    float tempfloat=0;
    float deltaR;
    word RawReading;
    // Serial.println (F("Got Here: FahrenheitEquals"));
    delay (2000);
    tempfloat=Serial.parseFloat();
    if (tempfloat != 0) CalculateDegreeOffset(tempfloat);
    else ValueNotAccepted();  
    DrainCmdTermiantors();
  }

//---------------------------------------------------------------------- 
void CelsiusEquals()
 { // sets offset according to current reading and input Fahrenheit
    float tempfloat=0;
    float deltaR;
    word RawReading;
    // Serial.println (F("Got Here: CelsiusEquals"));
    delay (2000);
    tempfloat=Serial.parseFloat();
    if (tempfloat != 0) CalculateDegreeOffset((tempfloat*1.8)+32);
    else ValueNotAccepted();  
    DrainCmdTermiantors();
  }

//---------------------------------------------------------------------- 
void NewRefVolt()
  { // New Degree Offset
    float tempfloat=0;
    delay (2000);
    tempfloat=Serial.parseFloat();
    if (tempfloat!=0)
      { // Serial.println(tempfloat,4);
        // Serial.println(RefVoltage,4);
        RefVoltage=tempfloat;
        // Serial.println(RefVoltage,4);
        newflg = newflg |  B00000100;
        PrintOKStr();
      }
    else ValueNotAccepted();    
    DrainCmdTermiantors();
  }

//-----------------------------------------------------
void RestoreFromBackup()
  { char TempString[EEwdsize];
    byte i;
    // read the backup copy
    for (i=0; i<EEwdsize; i++) TempString[i]=EEPROM.read(StorageBackup + i);
    // write working copy
    for (i=0; i<EEwdsize; i++) EEPROM.write(StorageWorking +i, TempString[i]);
    newflg=0;
    Read_Calibration_Data();
    Report_Reset();
  }

//---------------------------------------------------------------------- 
void OverwriteBackup()
  { char TempString[EEwdsize];
    byte i;
    // read the working copy
    for (i=0; i<EEwdsize; i++) 
	         TempString[i]=EEPROM.read(StorageWorking + i);
    // write backup copy
    for (i=0; i<EEwdsize; i++) EEPROM.write(StorageBackup +i, TempString[i]);
    PrintOKStr();
  }

//=====================================================================
void TestData1()
 { // These sets were picked for testing
   // so that one set look like the another set.
   char temp[]="(1)tst data,Nano   ";
   //...........1234567890123456
   byte i;
   Serial.println(F("; Test Data one being written to EEPROM"));
   // clear the EEPROM report storage area
   ClearStorage();
   for (i=0; i< EEidsize; i++) IdString[i]=temp[i];
   // Insert null terminator at end
   IdString[EEidsize]=0;
   DegreeOffset=0;
   RefVoltage  =1.075;
   MinuteTarget=1;
   newflg=0XFF;
   Write_Calibration_Data();
 }

//-----------------------------------------------------
void TestData2()
 { char temp[]="(2)tst data, UNO   ";
   //...........1234567890123456
   byte i;
   Serial.println(F("; Test Data two being written to EEPROM"));
   // clear the EEPROM report storage area
   ClearStorage();
   for (i=0; i< EEidsize; i++) IdString[i]=temp[i];
   // Insert null terminator at end
   IdString[EEidsize]=0;
   DegreeOffset=0;
   RefVoltage  =1.1000;
   MinuteTarget=1;
   newflg=0XFF;
   Write_Calibration_Data();
}

//-----------------------------------------------------
void CalibrationMode()
  { // used for calbration, reduces time between report lines to 5 seconds
    // there should be about 40 samples per report which will still give a good average
    
    if (SecondsMinute==5000)
      {
      Serial.println(F("; Exiting 5 second calibration mode <<<"));
      SecondsMinute=60000;
      }
    else
     {
     Serial.println(F("; Entering 5 second calibration mode >>>"));
     SecondsMinute=5000;
     }
    MinuteTarget=1;
    DeBug=false;
    EepromMode=false;
    ReportMode=true;
    RtnRawRead = true;
    RtnCelsius = true; 
    RtnFahrenh = true;
    Report_Reset();
  }

//-----------------------------------------------------
void EepromDumpAll()
  { char buffer[60];                                       // allocate buffer 
    word addr=0;                                           // set start address
    PrintSeperatorLine();
    Serial.println(F("; Dump all EEPROM in Hex and ASCII"));  // inform the user
    while (addr < E2END)                                   // run until we reach the end
      { for (byte i=0; i<16; i++)                          // process 16 bytes at a time
          { buffer[i]=EEPROM.read(addr++);                 // read EEPROM 
          }
        Serial.print ("; ");  
        Serial.println(formatRamDump(addr-16, buffer));    // print formatted string
      }
    PrintSeperatorLine();  
    PrintOKStr();   
  }
// =================== end of thermometer functions code ====================== //
//
// ========= Previous defined functions INVALID in this implimentation ======== // 
//void CelsiusEquals()
// { // sets offset according to current reading and input Celsius
//    float tempfloat=0;
//    word deltaR=0;
//    // Serial.println (F("Got Here: CelsiusEquals"));
//    delay (2000);
//    tempfloat=Serial.parseFloat();
//    if (tempfloat != 0) 
//      { // get the current raw reading
//        Serial.println (F("; Calculating new offset ... ..."));
//        deltaR = (tempfloat - DegreeOffset) / RefVoltage;
//        while (CycleCount<250) avrRawTemp();
//        CovrtOffsetR = (Accumalator/CycleCount)-deltaR;
//        newflg = newflg |  B00000010;
//        PrintOKStr();
//      }
//    else Serial.println (F("; C= zero value not accepted"));  
//    DrainCmdTermiantors();
//  }
//
//---------------------------------------------------------------------- 
//void FahrenheitEquals()
// { // sets offset according to current reading and input Fahrenheit
//    float tempfloat=0;
//    float deltaR;
//    // Serial.println (F("Got Here: FahrenheitEquals"));
//    delay (2000);
//    tempfloat=Serial.parseFloat();
//    if (tempfloat != 0) 
//      { // get the current raw reading
//        Serial.println (F("; Calculating new offset ... ..."));
//        deltaR = (tempfloat - CovrtOffsetF) / Covrt2Fahrenheit;
//        while (CycleCount<250) avrRawTemp();
//        CovrtOffsetR = (Accumalator/CycleCount)-deltaR;
//        newflg = newflg |  B00000010;
//        PrintOKStr();
//      }
//    else Serial.println (F("; F: zero value not accepted"));  
//    DrainCmdTermiantors();
//  }
//
//---------------------------------------------------------------------- 
//void NewOffsetR()
//  { // New Conversion Offset
//    int tempfloat=0;
//    delay (2000);
//    tempfloat=Serial.parseFloat();
//    if (tempfloat!=0)
//      { CovrtOffsetR=tempfloat;
//        newflg = newflg |  B00000010;
//        PrintOKStr();
//      }
//    else Serial.println (F("; O: zero value not accepted"));    
//    DrainCmdTermiantors();
//  }
//  
//---------------------------------------------------------------------- 
//void NewCelsius()
//  { // New Celsius Factor
//    delay (2000);
//    float tempfloat=0;
//    // Serial.println ("Got Here: NewCelsius");
//    tempfloat=Serial.parseFloat();
//    if (tempfloat != 0) 
//      { RefVoltage=tempfloat;
//        newflg = newflg |  B00000100;
//        // Calculate Fahrenheit factor
//        Covrt2Fahrenheit=RefVoltage * 1.8000;
//        PrintOKStr();
//      }
//    else Serial.println (F("; C: zero value not accepted"));  
//    DrainCmdTermiantors();
//  }
//                 
//---------------------------------------------------------------------- 
//void NewFahrenheit()
//  { // New Fahrenheit constant
//    delay (2000);
//    float tempfloat=0;
//    // Serial.println ("Got Here: NewFahrenheit");
//    tempfloat=Serial.parseFloat();
//    if (tempfloat != 0) 
//      { Covrt2Fahrenheit=tempfloat;
//        newflg = newflg |  B00000100;
//        // Calculate Celsius factor
//        RefVoltage = Covrt2Fahrenheit / 1.8000;
//        PrintOKStr();
//      }
//    else Serial.println (F("; F= zero value not accepted"));  
//    DrainCmdTermiantors();
//  }
//  
//=========================================================================================
//      ---- Below is the version for the internal Temperature Sensor ----
//-----------------------------------------------------------------------------------------
//void avrRawTemp()
//  { /* each sample has 16 ADC reads for a 12 bit virtual ADC       */
//    /* REF: Atmel document number AVR121.pdf                       */
//    //-------------------------------------------------------------//
//    /* on 16Mhz ATmega328 512 samples requires just under 1 second */
//    /* 16 samples (16*16=256) gives fairly consistent results      */
//    /*    on a steady-state system in under 40K microseconds       */
//    //-------------------------------------------------------------//
//    unsigned long RawSum=0;                         // used to sum samples for averaging 
//    word RawTemp=0;                                 // used to accumalate 10 bit ADC readings 
//    word test=0;                                    // used to count samples
//    byte k;                                         // counter for 10 bit ADC reads
//    
//    // turn on internal reference, right-shift ADC buffer,ADC channel = internal temp sensor
//    // B11000000: AMUX Voltage reference  = Internal 1.1 volt        (bits 7,6)
//    // B00000000: AMUX Right Shift ADC Buffer                        (bit 5)
//    // B00001000: AMUX Input Source = Internal Temperature Sensor    (bits 3,2,1,0)
//    ADMUX = B11000001;
//    delay(5);    			            // wait a bit for the analog ref to stabilize
//    
//    while (test++ < 64 )                            // oversampling loop (for averaging) 
//          {   for (k=0; k<16; k++) 	           // virtual ADC loop, 
//                                                    // 16 consecutiive readings
//                { ADCSRA |= _BV(ADSC); 	           // start the conversion                                
//                  while (bit_is_set(ADCSRA, ADSC)); // ADSC is cleared when the 
//                                                    // conversion finishes
//                  RawTemp += (ADCL | (ADCH << 8));  // accumalate the reading (low byte first)
//                }
//              RawSum += (RawTemp >>2);              // accumalate virtual 12 bit ADC value
//              RawTemp=0;                            // zero ADC accumalator for 
//                                                    // the next sequence
//          }
//    // Serial.println ("Got here: avrRawTemp");
//    Accumalator += ((RawSum)>>6);                   // averag by shifting bit position, 
//                                                    // LSBs are lost 
//    CycleCount++;                                   // used by functions to average readings 
//  }
//-----------------------------------------------------------------------------------------
//      ---- This is the version for the internal Temperature Sensor ----
//-----------------------------------------------------------------------------------------
//  
//void Convert(word RawReading)
//  { // converts Raw Reading to Celsius and Fahrenheit
//    // we need to use an interger for the delta to allow for negative values
//    int Delta;
//    
//    // get the difference between the current reading and the raw reading offset
//    Delta=RawReading-CovrtOffsetR;
//    
//    // Serial.println (F("; got here: Convert"));
//    // Serial.println (RawReading);
//    // Serial.println (CovrtOffsetR);
//    // Serial.println (Delta);
//    // Serial.println (RefVoltage);
//    // Serial.println (DegreeOffset);
//    
//    Celsius= (Delta * RefVoltage) + DegreeOffset;
//    // Serial.println (Celsius);
//    Celsius=nearestquater(Celsius);
//    
//    Fahrenheit= (Delta * Covrt2Fahrenheit) + CovrtOffsetF;
//    // Serial.println (Fahrenheit);
//    Fahrenheit= nearesthalf(Fahrenheit);
//  }
//-----------------------------------------------------------------------------------------
//      ---- This is the version for the internal Temperature Sensor ----

